#include <atari.h>
#include "atari_multijoy.h"

/* FIXME: Will be part of atari.h's Page 2 labels */
#define VDSLST (*(unsigned int*)0x200)
#define SDMCTL (*(unsigned char*)0x22F)


/**
 * Configuring controller ports (#1 and #2, aka combined PORTA)
 * for use with a Multijoy system (input from port 1, output to port 2),
 * or back to default (input from both ports).
 * 
 * @param bool multi - if true, configure for multijoy; else normal
 */
void multijoy_init(bool multi) {
  PIA.pactl = 0x00;
  if (multi) {
    PIA.porta = 0xF0; /* 4 higher bits = out, 4 lower bits = in */
  } else {
    PIA.porta = 0xF0; /* all 8 bits = in */
  }
  PIA.pactl = PxCTL_ADDRESSING | PACTL_MOTOR_CONTROL | 0x30;
}


/**
 * multijoy_read() - Immediately reads the stick & trigger state of the
 * specified multijoy joystick controller, by first setting the
 * joystick on the output port, waiting a few cycles, the reading
 * the values from the input port, and returning them.
 *
 * @param unsigned char stick - the joystick to read (0-7)
 * @return unsigned char - the combined state of the stick (low 4 bits)
 *   and trigger (5th bit); e.g. 0xF (dec. 15) plus 0x10 (dec. 16),
 *   or 0x1F (dec. 31) for a joystick that's centered, with the trigger
 *   released.
 */
#pragma optimize (push, off)
unsigned char multijoy_read(unsigned char stick) {
  unsigned char s, f;

  stick <<= 4;
  PIA.porta = stick;

  // Delay 30 cycles before reading PORTA
  asm("ldx #$06");
WAI:
  asm("dex");
  asm("bne %g", WAI);

  s = PIA.porta & 0x0F;
  f = GTIA_READ.trig0;

  return s | (f << 4);
}
#pragma optimize (pop)


/**
 * multijoy_dli_1() & multijoy_dli_2() -
 * A pair of Display List Interrupt (DLI) handlers that
 * constantly read the state of all 8 joysticks, and stores all
 * of their values in the first 16 bytes of Page 6 (0x600);
 * first 8 bytes for each joystick;s direction, second 8 bytes for each
 * joystick's firebutton.
 * Each phase 1 sets the high bits of PORTA to tell the MultiJoy which stick we want;
 * each phase 2 reads the stick & trigger values, and records them.
 */

/* N.B. The DLIs don't touch the Y register, so we don't bother pushing 
   it to / popping it from the stack */
#define DLI_START asm("pha"); asm("txa"); asm("pha"); /* asm("tya"); asm("pha"); */
#define DLI_END /* asm("pla"); asm("tay"); */ asm("pla"); asm("tax"); asm("pla"); asm("rti");

#pragma optimize (push, off)

unsigned char multijoy_dli_stick = 0;

void multijoy_dli_2(void);

void multijoy_dli_1(void) {
  DLI_START

  /* ANTIC.wsync = 0; */
  asm("lda #0");
  asm("sta $d40a"); /* WSYNC */

#ifdef MULTIJOY_VISUAL_DEBUG
  /* GITA_WRITE.colbk = ANTIC.vcount; */
  asm("lda $d40b"); /* VCOUNT */
  asm("sta $d01a"); /* COLBK */
#endif

  /* PIA.porta = multijoy_dli_stick << 4; */
  asm("lda %v", multijoy_dli_stick);
  asm("asl");
  asm("asl");
  asm("asl");
  asm("asl");
  asm("sta $d300"); /* PORTA */

  /* VDLST = multijoy_dli_2; */
  asm("lda #<%v", multijoy_dli_2);
  asm("ldx #>%v", multijoy_dli_2);
  asm("sta $200"); /* VDSLST (lo) */
  asm("stx $201"); /* VDSLST (hi) */

  DLI_END
}

void multijoy_dli_2(void) {
  DLI_START

  /* ANTIC.wsync = 0; */
  asm("lda #0");
  asm("sta $d40a"); /* WSYNC */

#ifdef MULTIJOY_VISUAL_DEBUG
  /* GITA_WRITE.colbk = 0; */
  asm("lda #$0");
  asm("sta $d01a"); /* COLBK */
#endif

  asm("ldx %v", multijoy_dli_stick);

  /* POKE(600 + multijoy_dli_stick, (PIA.porta & 0xf)); */
  asm("lda $d300"); /* PORTA */
  asm("and #$0f");
  asm("sta $600,x");

  /* POKE(608 + multijoy_dli_stick, GTIA_READ.trig0); */
  asm("lda $d010"); /* TRIG0 */
  asm("sta $608,x");

  /* multijoy_dli_stick = (multijoy_dli_stick + 1) & 7; */
  asm("inx");
  asm("txa");
  asm("and #$07");
  asm("sta %v", multijoy_dli_stick);

  /* VDLST = multijoy_dli_1; */
  asm("lda #<%v", multijoy_dli_1);
  asm("ldx #>%v", multijoy_dli_1);
  asm("sta $200"); /* VDSLST (lo) */
  asm("stx $201"); /* VDSLST (hi) */

  DLI_END
}

#pragma optimize (pop)


/**
 * multijoy_dli_init_1() - Prepare for caller to set up Display List Interrupt bit on
 * some mode lines.  (Make sure DLI interrupt is disabled, blanks screen,
 * waits one screen refresh.)
 *
 * @return The value of the SDMCTL register prior to screen being blanked
 *   (feed it to multijoy_dli_init_2(), which will return it to that value when
 *   the DLI setup process is completed)
 */
unsigned char multijoy_dli_init_1(void) {
  unsigned char sdmctl;

  ANTIC.nmien = NMIEN_VBI;

  sdmctl = SDMCTL;
  SDMCTL = DMACTL_PLAYFIELD_NONE;

  while (ANTIC.vcount < 124) {
    /* wait */
  }

  return sdmctl;
}

/**
 * multijoy_dli_init_2() - Complete Display List Interrupt setup, to be done
 * after the caller has set the DLI bit on some mode lines.  Enables DLIs,
 * starting with the phase 1 routine above, waits a moment, and reenables
 * the screen.
 *
 * @param unsigned char sdmctl - The value to place in the SDMCTL register after
 *   the DLI setup process is complete. (Probably the return value you got back
 *   from multijoy_dli_init_1(), above.)
 */
void multijoy_dli_init_2(unsigned char sdmctl) {
  VDSLST = (unsigned) multijoy_dli_1;
  ANTIC.nmien = NMIEN_VBI | NMIEN_DLI;

  while (ANTIC.vcount < 124) {
    /* wait */
  }

  SDMCTL = sdmctl;
}


/**
 * multijoy_wait() - Wait until we've read all 8 joysticks.
 * (Does this by waiting in a tight loop to see the internal stick
 * counter climb to 7.)
 */
void multijoy_wait(void) {
  while (multijoy_dli_stick < 7) {
    /* wait */
  }
}
